/// 2D linear regression plugin implemented from 
/// https://gist.github.com/tansey/1375526
/// id144.org

#region usings
using System;
using System.ComponentModel.Composition;

using VVVV.PluginInterfaces.V1;
using VVVV.PluginInterfaces.V2;
using VVVV.Utils.VColor;
using VVVV.Utils.VMath;

using VVVV.Core.Logging;
#endregion usings

namespace VVVV.Nodes
{
	#region PluginInfo
	[PluginInfo(Name = "LinearRegression", Category = "Value", Help = "2D linear regression", Tags = "", Author="id144")]
	#endregion PluginInfo
	public class ValueLinearRegressionNode : IPluginEvaluate
	{
		#region fields & pins
		[Input("Input XY", DefaultValue = 1.0)]
		public ISpread<double> FInput;
		
		[Input("Inclusive Start", DefaultValue = 1.0)]
		public ISpread<int> FInputInclusiveStart;

		[Input("Exclusive End", DefaultValue = 1.0)]
		public ISpread<int> FInputExclusiveEnd;

		[Output("Output")]
		public ISpread<double> FOutput;
		
		[Output("R squared")]
		public ISpread<double> FOutputRSquared;
		
		[Output("Y intercept")]
		public ISpread<double> FOutputYIntercept;	
		
		[Output("Slope")]
		public ISpread<double> FOutputSlope;
		
		[Import()]
		public ILogger FLogger;
		#endregion fields & pins
        /// <summary>
        /// Fits a line to a collection of (x,y) points.
        /// </summary>
        /// <param name="xVals">The x-axis values.</param>
        /// <param name="yVals">The y-axis values.</param>
        /// <param name="inclusiveStart">The inclusive inclusiveStart index.</param>
        /// <param name="exclusiveEnd">The exclusive exclusiveEnd index.</param>
        /// <param name="rsquared">The r^2 value of the line.</param>
        /// <param name="yintercept">The y-intercept value of the line (i.e. y = ax + b, yintercept is b).</param>
        /// <param name="slope">The slop of the line (i.e. y = ax + b, slope is a).</param>

        public static void LinearRegression(double[] xVals, double[] yVals, 
                                            int inclusiveStart, int exclusiveEnd, 
                                            out double rsquared, out double yintercept, 
                                            out double slope)
        {
           // Debug.Assert(xVals.Length == yVals.Length);
            double sumOfX = 0;
            double sumOfY = 0;
            double sumOfXSq = 0;
            double sumOfYSq = 0;
            double ssX = 0;
            double ssY = 0;
            double sumCodeviates = 0;
            double sCo = 0;
            double count = exclusiveEnd - inclusiveStart;

            for (int ctr = inclusiveStart; ctr < exclusiveEnd; ctr++)
            {
                double x = xVals[ctr];
                double y = yVals[ctr];
                sumCodeviates += x * y;
                sumOfX += x;
                sumOfY += y;
                sumOfXSq += x * x;
                sumOfYSq += y * y;
            }
            ssX = sumOfXSq - ((sumOfX * sumOfX) / count);
            ssY = sumOfYSq - ((sumOfY * sumOfY) / count);
            double RNumerator = (count * sumCodeviates) - (sumOfX * sumOfY);
            double RDenom = (count * sumOfXSq - (sumOfX * sumOfX))
             * (count * sumOfYSq - (sumOfY * sumOfY));
            sCo = sumCodeviates - ((sumOfX * sumOfY) / count);

            double meanX = sumOfX / count;
            double meanY = sumOfY / count;
            double dblR = RNumerator / Math.Sqrt(RDenom);
            rsquared = dblR * dblR;
            yintercept = meanY - ((sCo / ssX) * meanX);
            slope = sCo / ssX;
        }
		//called when data for any output pin is requested
		public void Evaluate(int SpreadMax)
		{
			FOutput.SliceCount = 1;
			FOutputRSquared.SliceCount = 1;
			FOutputSlope.SliceCount = 1;
			FOutputYIntercept.SliceCount = 1;
			
			double[] xVals = new double[FInput.SliceCount/2];
			double[] yVals = new double[FInput.SliceCount/2];
			for (int i = 0; i < (FInput.SliceCount/2); i++)
			{
				xVals[i] = FInput[i*2];				
				yVals[i] = FInput[i*2+1];
			}
			double _rSquared;
			double _yIntercept;
			double _slope;
			LinearRegression(xVals,yVals, FInputInclusiveStart[0], FInputExclusiveEnd[0], out _rSquared, out _yIntercept, out _slope);
			
			FOutputRSquared[0] = _rSquared;
			FOutputYIntercept[0] = _yIntercept;		
			FOutputSlope[0] = _slope;
		}
	}
}
